/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include"engine.h"
#include<fstream>
#include<ctime>
#include<iostream>
#include<math.h>
#include <libxml/parser.h>
#include"globals.h"
#include"defaults.h"
#include"utils.h"

PCalEngine::PCalEngine(int in_bufferDays,
                   int in_badFeelDaysBefore,
                   int in_badFeelDaysAfter,
                   int in_regularMinLength,
                   int in_regularMaxLength):
bufferDays(in_bufferDays),
badFeelDaysBefore(in_badFeelDaysBefore),
badFeelDaysAfter(in_badFeelDaysAfter),
regularMinLength(in_regularMinLength),
regularMaxLength(in_regularMaxLength),
modified(false)
{
    // compress the XML output by default
    xmlSetCompressMode(9);
    recalculate();
}

void PCalEngine::changeOptions(int in_bufferDays,
                   int in_badFeelDaysBefore,
                   int in_badFeelDaysAfter,
                   int in_regularMinLength,
                   int in_regularMaxLength)
{
    if((bufferDays != in_bufferDays) ||
       (badFeelDaysBefore != in_badFeelDaysBefore) ||
       (badFeelDaysAfter != in_badFeelDaysAfter) ||
       (regularMinLength != in_regularMinLength) ||
       (regularMaxLength != in_regularMaxLength))
    {
        bufferDays = in_bufferDays;
        badFeelDaysBefore = in_badFeelDaysBefore;
        badFeelDaysAfter = in_badFeelDaysAfter;
        regularMinLength = in_regularMinLength;
        regularMaxLength = in_regularMaxLength;
        recalculate();
    }
}

PCalDayInfo PCalEngine::getDayInfo(const Glib::Date date)
{
    PCalDayInfo info;
    int days_between = 0;
    std::map<Glib::Date, PCalRecordType>::const_iterator p;
    
    if(!start_dates.size())
    {
        // no data at all
        info.day_num = -1;
    }
    else 
    if((p = start_dates.find(date)) != start_dates.end()) {
        // this date is really in the database as a day 1
        info.day_num = 0;
        info.fertile = false;
        info.pregnancy = (p->second == PCAL_TYPE_PREGNANCY);
        info.birth = (p->second == PCAL_TYPE_BIRTH);
        info.bad_feel = true;
        info.estimate = false;
    }
    else
    if(!periodsCount)
    {
        // no data at all
        info.day_num = -1;
    }
    else 
    if(start_dates.begin()->first > date)
    {
        // the requested date is before the recorded periods
        info.day_num = -1;
    }
    else { 
        std::map<Glib::Date, PCalRecordType>::const_iterator last_record = --start_dates.end();
        // is the date in the recorded period?
        if(last_record ->first >= date)
        {
            // find the starting date of the date's period
            p = start_dates.lower_bound(date);
            if(p!=start_dates.end()) {                
                // this is a date in the complete period
                // this is not a day 1 - we have already checked this
                days_between = date.days_between(p->first);
                info.day_num = (--p)->first.days_between(date);
                int period_length = info.day_num + days_between;
                if(p->second == PCAL_TYPE_MENS) {                    
                    // check if this is a regular period
                    if((period_length < regularMinLength) ||
                       (period_length > regularMaxLength)) {
                        info.day_num = -1;
                    } else {
                        info.fertile = (days_between < 18) && (days_between > 11);
                        info.bad_feel = (info.day_num<=badFeelDaysAfter)||(days_between<=badFeelDaysBefore);
                    }
                    info.pregnancy = false;
                    info.birth = false;                    
                } else 
                if(p->second == PCAL_TYPE_PREGNANCY) {
                    if(info.day_num <= REGULAR_PREGNANCY_DAYS) {
                        info.fertile = false;
                        info.pregnancy = true;
                        info.birth = false;
                        info.bad_feel = true;
                    } else {
                        // this is a day after the normal pregnancy period
                        // know nothing
                        info.day_num = -1;
                    }
                } else {
                    // know nothing
                    info.day_num = -1;
                }
            } else {
                // something wrong - this should normally never happen
                info.day_num = -1;
            }
            info.estimate = false; 
        } else {
            // the day parameters should be predicted
            info.day_num = last_record->first.days_between(date);
            if(last_record->second == PCAL_TYPE_MENS) {
                // the cycle number is calculated by the average cycle length but
                // the fertility will be 
                if(avgLength)
                info.day_num %= avgLength;
    
                if(!info.day_num) {
                    // the most probably this will be a day 1
                    info.fertile = false;
                    info.bad_feel = true;
                } else {
                    // this is a day after the last recorded period start
                    info.fertile = (info.day_num > (minLength - 18 - bufferDays)) 
                                   && (info.day_num < (maxLength - 11 + bufferDays));
                    info.bad_feel = (info.day_num<=badFeelDaysAfter)||
                                    ((avgLength-info.day_num)<=badFeelDaysBefore);
                }
                info.pregnancy = false;
                info.birth = false;                    
            } else 
            if(last_record->second == PCAL_TYPE_PREGNANCY) {
                if(info.day_num < REGULAR_PREGNANCY_DAYS) {
                    info.fertile = false;
                    info.pregnancy = true;
                    info.birth = false;
                    info.bad_feel = true;
                } else
                if(info.day_num == REGULAR_PREGNANCY_DAYS) {
                    info.fertile = false;
                    info.pregnancy = false;
                    info.birth = true;
                    info.bad_feel = true;
                } else {
                    // this is a day after the normal pregnancy period
                    // know nothing
                    info.day_num = -1;
                }
            } else {
                // know nothing
                info.day_num = -1;
            }
            info.estimate = true; 
        }
    }
    return info;
}

bool PCalEngine::loadFromFile(const Glib::ustring fname)
{
    Glib::ustring buff;
    Glib::Date tmp_date;

    xmlDocPtr doc;
    xmlNodePtr rootptr;
    xmlNodePtr versionPtr = 0;
    xmlNodePtr dataPtr = 0;
    xmlNodePtr cur;

    start_dates.clear();
    notes_container.clear();

    // for some reason glib/glibmm crashed unless the temporary 
    //variable is used
    Glib::ustring bloodyGlib = Glib::filename_from_utf8(fname);
    doc = xmlParseFile(bloodyGlib.c_str());

    if (doc == NULL ) {
        return false;
    }

	rootptr = xmlDocGetRootElement(doc);

	if (rootptr == NULL) {
        // file is empty
		xmlFreeDoc(doc);
		return false;
	}

	if (xmlStrcmp(rootptr->name, BAD_CAST "pcalendar")) {
		// document of the wrong type
		xmlFreeDoc(doc);
		return false;
	}
    // get the root children
    for (cur = rootptr->children; cur; cur = cur->next) {
        if((cur->type == XML_ELEMENT_NODE)) {
            if(xmlStrcmp(cur->name, BAD_CAST "version")==0) {
                versionPtr = cur;
            } else
            if(xmlStrcmp(cur->name, BAD_CAST "data")==0) {
                dataPtr = cur;
            } 
        }
    }

    // check the document format version    
    if(!versionPtr) {
        xmlFreeDoc(doc);
        return false;
    }

    if(xmlStrcmp(versionPtr->children->content, BAD_CAST "2.0")) {
        // wrong file format version
        xmlFreeDoc(doc);
        return false;
    }

    // load the data
    if(dataPtr) {
        for (cur = dataPtr->children; cur; cur = cur->next) {
            if((cur->type == XML_ELEMENT_NODE)) {
                /* Add the menstruation record */
                if(xmlStrcmp(cur->name, BAD_CAST "mens")==0) {
                    if(cur->children) {
                        buff = Glib::ustring((char *)cur->children->content);
                    }
                    if(buff=="") continue;
                    if(!PCal::stringToDate(buff, tmp_date)) {
                        // wrong file format
                        start_dates.clear();
                        modified = false;
                    } else {
                        start_dates[tmp_date]=PCAL_TYPE_MENS;
                    }
                } else
                /* Add the pregnancy record */
                if(xmlStrcmp(cur->name, BAD_CAST "pregnancy")==0) {
                    if(cur->children) {                    
                        buff = Glib::ustring((char *)cur->children->content);
                    }
                    if(buff=="") continue;
                    if(!PCal::stringToDate(buff, tmp_date)) {
                        // wrong file format
                        start_dates.clear();
                        modified = false;
                    } else {
                        start_dates[tmp_date]=PCAL_TYPE_PREGNANCY;
                    }
                } else
                /* Add the birth record */
                if(xmlStrcmp(cur->name, BAD_CAST "birth")==0) {
                    if(cur->children) {
                        buff = Glib::ustring((char *)cur->children->content);
                    }
                    if(buff=="") continue;
                    if(!PCal::stringToDate(buff, tmp_date)) {
                        // wrong file format
                        start_dates.clear();
                        modified = false;
                    } else {
                        start_dates[tmp_date]=PCAL_TYPE_BIRTH;
                    }
                } else
                /* Load the date note */
                if(xmlStrcmp(cur->name, BAD_CAST "note")==0) {
                    Glib::Date note_date;
                    Glib::ustring note_text = "";
                    bool wrong_date = false;
                    for (xmlNodePtr noteN = cur->children; noteN; noteN = noteN->next) {
                        if(noteN->type == XML_ELEMENT_NODE) {
                            if(xmlStrcmp(noteN->name, BAD_CAST "date")==0) {
                                if(noteN->children) {
                                    buff = Glib::ustring((char *)noteN->children->content);
                                }
                                if(buff=="") {
                                    wrong_date = true;
                                    break;
                                }
                                if(!PCal::stringToDate(buff, note_date)) {
                                    wrong_date = true;
                                    break;
                                }                                
                            } else
                            if(xmlStrcmp(noteN->name, BAD_CAST "text")==0) {
                                if(noteN->children) {
                                    note_text = Glib::ustring((char *)noteN->children->content);
                                }
                            }
                        }
                    }
                    // at this point the date and note should be filled
                    if(!wrong_date) {
                        notes_container[note_date] = note_text;
                    } else {
                        // having the damaged note is not critical
                        wrong_date = false;
                    }
                }
            }
        }
    }
    xmlFreeDoc(doc);

    modified = false;
    recalculate();

    return true;
}

bool PCalEngine::saveToFile(const Glib::ustring fname)
{
    // This is a libxml2 stuff
    xmlDocPtr doc;
    xmlNodePtr n;
    xmlNodePtr note_root;
    
    // set the compression mode to max
    doc = xmlNewDoc(BAD_CAST "1.0");
    n = xmlNewNode(NULL, BAD_CAST "pcalendar");    
    xmlDocSetRootElement(doc, n);
    xmlNewChild(n, NULL, BAD_CAST "version", BAD_CAST "2.0");
    n = xmlNewChild(n, NULL, BAD_CAST "data", NULL);

    // dates
    for(std::map<Glib::Date, PCalRecordType>::const_iterator p = start_dates.begin();
        p!=start_dates.end(); 
        p++)
    {
        if(p->second == PCAL_TYPE_MENS) {
            xmlNewChild(n, NULL, 
                    BAD_CAST "mens", 
                    BAD_CAST p->first.format_string("%Y/%m/%d").c_str());
        } else
        if(p->second == PCAL_TYPE_PREGNANCY) {
            xmlNewChild(n, NULL, 
                    BAD_CAST "pregnancy", 
                    BAD_CAST p->first.format_string("%Y/%m/%d").c_str());
        } else
        if(p->second == PCAL_TYPE_BIRTH) {
            xmlNewChild(n, NULL, 
                    BAD_CAST "birth", 
                    BAD_CAST p->first.format_string("%Y/%m/%d").c_str());
        }
    }
    
    // save notes
    for(std::map<Glib::Date, Glib::ustring>::const_iterator pn = notes_container.begin();
        pn!=notes_container.end(); 
        pn++)
    {
        note_root = xmlNewChild(n, NULL, 
                BAD_CAST "note", 
                NULL);
        xmlNewChild(note_root, NULL, 
                BAD_CAST "date", 
                BAD_CAST pn->first.format_string("%Y/%m/%d").c_str());
        xmlNewChild(note_root, NULL, 
                BAD_CAST "text", 
                BAD_CAST pn->second.c_str());
    }
    // save the XML file
    Glib::ustring bloodyGlib = Glib::filename_from_utf8(fname);
    if(xmlSaveFormatFile (bloodyGlib.c_str(), 
       doc, 1) == -1) {
        xmlFreeDoc(doc);
        return false;
    }
    xmlFreeDoc(doc);
    
    modified = false;
    return true;
}

bool PCalEngine::importFromFile(const Glib::ustring fname, int file_format)
{
    if(file_format == 0) {
        Glib::ustring buff;
        Glib::Date d;
        
        Glib::ustring bloodyGlib = Glib::filename_from_utf8(fname);
        std::ifstream in_file(bloodyGlib.c_str());    
        if(!in_file) return false;
            
        // there is no dates parsing routine in the Glib
        while(!in_file.eof()) {
            in_file>>buff;       
            if(buff=="") continue;
        
            unsigned day, month, year;
            if(!sscanf(buff.c_str(), "%u/%u/%u", &year, &month, &day))
            {
                // file format error
                start_dates.clear();
                modified = false;
                return false;
            }
            if(d.valid_dmy(day,
                  (Glib::Date::Month)month,
                  year)) 
            {
                d.set_dmy(day,
                      (Glib::Date::Month)month,
                      year);
            } else {
                // file format error
                start_dates.clear();
                modified = false;
                return false;
            }
            start_dates[d]=PCAL_TYPE_MENS;
        }
    } // file_format == 0
    modified = true;
    recalculate();
    return true;
}

bool PCalEngine::exportToFile(const Glib::ustring fname, int file_format)
{
    if(file_format == 0) {
        Glib::ustring bloodyGlib = Glib::filename_from_utf8(fname);
        std::ofstream out_file(bloodyGlib.c_str());    
        if(!out_file) return false;
            
        for(std::map<Glib::Date, PCalRecordType>::const_iterator p = start_dates.begin();
            p!=start_dates.end(); 
            p++)
        {
            if(p->second == PCAL_TYPE_MENS) {
                out_file<<p->first.format_string("%Y/%m/%d")<<std::endl;
            }
        }    
    } // file_format == 0    
    return true;
}

void PCalEngine::addStartDate(const Glib::Date date)
{
    start_dates[date]=PCAL_TYPE_MENS;
    modified = true;
    recalculate();
}
void PCalEngine::addPregnancyDate(const Glib::Date date)
{
    start_dates[date]=PCAL_TYPE_PREGNANCY;
    modified = true;
    recalculate();
}
void PCalEngine::addBirthDate(const Glib::Date date)
{
    start_dates[date]=PCAL_TYPE_BIRTH;
    modified = true;
    recalculate();
}

void PCalEngine::removeDateRecord(const Glib::Date date)
{
    start_dates.erase(date);
    modified = true;
    recalculate();
}

void PCalEngine::clear()
{
    start_dates.clear();
    notes_container.clear();
    modified = false;
    recalculate();
}

void PCalEngine::recalculate()
{
    // at this point we have the sorted starting dates list
    periodsCount = 0;
    avgLength = 0;
    minLength = 0;
    maxLength = 0;
    int sumLength = 0;
    calMethodAccuracy = 3;
    if(start_dates.size()<2) {
        // there are no even a single period
        return;
    }
    
    // we need to calculate the average, min and max period length
    std::map<Glib::Date, PCalRecordType>::const_iterator pstart = start_dates.begin();
    std::map<Glib::Date, PCalRecordType>::const_iterator pend = pstart;
    for(pend++;
        pend!=start_dates.end(); 
        pend++)
    {
        // skip the non-menstruation neighbor records
        if((pstart->second != PCAL_TYPE_MENS)||
           (pend->second != PCAL_TYPE_MENS)) {
           pstart++;
           continue;
        }
        // pstart and pend points to starting and ending date of the period
        int p_length = pstart->first.days_between(pend->first);
           
        if((p_length < regularMinLength) ||
           (p_length > regularMaxLength)) {
            // bad period should not be taken into account
            
        } else {
            periodsCount++;
            if(p_length>maxLength) maxLength = p_length;
            if((p_length<minLength) || minLength == 0) minLength = p_length;
            sumLength+=p_length;        
        }
        pstart++;
    }
    if(periodsCount>0) avgLength = int(round(double(sumLength)/double(periodsCount)));

    // calendar method accuracy depends on the medium period length,
    // the difference between the longest and the shortest periodsCount
    // and the number of the recorded periods
    // 0 - high, 1 - moderate, 2 - low
    // for the high accuracy the conditions are almost unreal
    if((periodsCount>5) &&
       (avgLength>25) &&
       (avgLength<33) &&
       ((maxLength - minLength)<2))
      calMethodAccuracy = 0;
    else
    if((periodsCount<5)) calMethodAccuracy = 2;
    else
    if((maxLength - minLength)>3) calMethodAccuracy = 2;
    else
    if((avgLength<24) &&
       (avgLength>35)) calMethodAccuracy = 2;
    else calMethodAccuracy = 1;
}

const std::vector<PCalPeriodInfo> PCalEngine::getPeriodsStats()
{
    PCalPeriodInfo pinfo;
    std::vector<PCalPeriodInfo> result;

    if(start_dates.size()>1) {
        std::map<Glib::Date, PCalRecordType>::const_iterator pstart = start_dates.begin();
        std::map<Glib::Date, PCalRecordType>::const_iterator pend = pstart;
    
        for(pend++;
            pend!=start_dates.end(); 
            pend++)
        {
            // skip the non-menstruation neighbor records
            if((pstart->second != PCAL_TYPE_MENS)||
               (pend->second != PCAL_TYPE_MENS)) {
               pstart++;
               continue;
            }
            
            int p_length = pstart->first.days_between(pend->first);
     
            if((p_length < regularMinLength) ||
               (p_length > regularMaxLength)) {
                // bad period should not be taken into account            
            } else {
                pinfo.startDate = pstart->first;
                pinfo.endDate = pend->first;
                pinfo.length = p_length;
                result.push_back(pinfo);
            }
            pstart++;
        }
    }
   
    return result;
}

bool PCalEngine::getLastMenstruationDate(Glib::Date& result)
{
    if(start_dates.size()>0) {
        std::map<Glib::Date, PCalRecordType>::const_reverse_iterator p = start_dates.rbegin();
        if(p->second == PCAL_TYPE_MENS) {
            result = p->first;
            return true;
        }        
        return false;
    } else {
        return false;
    }
}

bool PCalEngine::getDateNote(const Glib::Date date, Glib::ustring& note)
{
    std::map<Glib::Date, Glib::ustring>::const_iterator p;
    if((p = notes_container.find(date)) != notes_container.end()) {
        note = p->second;
        return true;
    } else {
        return false;
    }
}

bool PCalEngine::existsDateNote(const Glib::Date date)
{
    return (notes_container.find(date) != notes_container.end());
}

void PCalEngine::setDateNote(const Glib::Date date, const Glib::ustring note)
{
    // need to check if any changes not to fall to the modified state
    // without a reason
    if(notes_container[date]!=note) {
        notes_container[date] = note;
        modified = true;
    }    
}

const std::map<Glib::Date, Glib::ustring>& PCalEngine::getDateNotes()
{
    return notes_container;
}

void PCalEngine::deleteDateNote(const Glib::Date date)
{
    notes_container.erase(date);
    modified = true;
}
